"""WHAD BLE GATT default services

This module provides different default GATT services as defined in
* https://www.bluetooth.com/specifications/specs/battery-service/
* https://www.bluetooth.com/specifications/specs/device-information-service/
"""
from whad.ble.profile.services.bas import BatteryService
from whad.ble.profile.services.dis import DeviceInformationService

__all__ = [
    "BatteryService",
    "DeviceInformationService"
]



"""This module provides a VirtuaDevice class that can be used with a WhadDeviceConnector
to interact with a device that doesn't support WHAD protocol. It allows to convert WHAD messages
to the corresponding specific API calls.

This class handles device connection, disconnection and read/write operations. All the
parsing magic is performed in our WhadDevice class.
"""
from .ubertooth import UbertoothDevice
from .rzusbstick import RZUSBStickDevice
from .apimote import APIMoteDevice
from .hci import HCIDevice
from .rfstorm import RFStormDevice
from .yard import YardStickOneDevice
from .pcap import PCAPDevice

__all__ = [
    "UbertoothDevice",
    "RZUSBStickDevice",
    "APIMoteDevice",
    "HCIDevice",
    "RFStormDevice",
    "YardStickOneDevice",
    "PCAPDevice"
]


"""
WHAD Enhanced ShockBurst base connector

This module provides the `ESB` connector class to interact with Enhanced
ShockBurst compatible devices.
"""
from scapy.packet import Packet

from whad.hub.discovery import Domain, Capability
from whad.device import WhadDeviceConnector
from whad.esb.esbaddr import ESBAddress
from whad.hub.esb import ESBMetadata
from whad.scapy.layers.esb import ESB_Hdr,ESB_Payload_Hdr,ESB_Ack_Response
from whad.helpers import message_filter
from whad.hub.generic.cmdresult import Success, CommandResult
from whad.exceptions import UnsupportedDomain, UnsupportedCapability
from whad.hub.esb import EsbNodeAddress, Commands
from whad.hub.events import JammedEvt


class ESB(WhadDeviceConnector):
    """
    Enhanced ShockBurst protocol connector.

    This connector drives a Enhanced ShockBurst (ESB) capable device with
    ESB-specific WHAD messages. It is required by various role classes to
    interact with a real device and pre-process domain-specific messages.
    """

    domain = "esb"

    def format(self, packet):
        """
        Converts a scapy packet with its metadata to a tuple containing a scapy packet with
        the appropriate header and the timestamp in microseconds.
        """
        if isinstance(packet, bytes):
            packet = ESB_Hdr(packet)
        return self.hub.esb.format(packet)

    def __init__(self, device=None, synchronous=False):
        """
        Initialize the connector, open the device (if not already opened), discover
        the services (if not already discovered).

        If `synchronous` is set to True, PDUs must be processed manually and
        won't be forwarded to PDU-related callbacks.
        """
        self.__ready = False
        super().__init__(device)

        # Metadata cache
        self.__cached_channel = None
        self.__cached_address = None

        # Capability cache
        self.__can_send = None
        self.__can_send_raw = None

        # Open device and make sure it is compatible
        self.device.open()
        self.device.discover()

        # Check if device supports Enhanced ShockBurst
        if not self.device.has_domain(Domain.Esb):
            raise UnsupportedDomain("ESB")

        # Mark connector as ready
        self.__ready = True

        # Set synchronous mode
        self.enable_synchronous(synchronous)

    def close(self):
        """Stop the WHAD adpater and close its related device.
        """
        self.stop()
        self.device.close()

    def can_sniff(self):
        """
        Determine if the device implements a sniffer mode.
        """
        commands = self.device.get_domain_commands(Domain.Esb)
        return (
            (commands & (1 << Commands.Sniff)) > 0 and
            (commands & (1 << Commands.Start))>0 and
            (commands & (1 << Commands.Stop))>0
        )

    def can_send(self):
        """
        Determine if the device can transmit packets.
        """
        if self.__can_send is None:
            commands = self.device.get_domain_commands(Domain.Esb)
            self.__can_send = ((commands & (1 << Commands.Send))>0 or
                               (commands & (1 << Commands.SendRaw)))
        return self.__can_send

    def send(self,pdu, address=None, channel=None, retransmission_count=1):
        """
        Send Enhanced ShockBurst packets (on a single channel).
        """
        if self.can_send():
            # If we don't have address or channels, use the cached ones
            tx_address = address if address is not None else self.__cached_address
            tx_channel = channel if channel is not None else self.__cached_channel
            if self.support_raw_pdu():
                # If we support raw PDU but only got a payload, build a packet
                if ESB_Hdr not in pdu:
                    packet = ESB_Hdr(address=tx_address) / pdu
                else:
                    packet = pdu
            # if we don't support raw PDU and got a packet, crop to keep only the payload
            elif ESB_Hdr in pdu:
                packet = pdu[ESB_Payload_Hdr:]
            # if we don't support raw PDU and got a payload, keep it as it is
            else:
                packet = pdu

            # Generate TX metadata
            packet.metadata = ESBMetadata()
            packet.metadata.channel = tx_channel
            packet.metadata.address = tx_address
            packet.metadata.retransmission_count = retransmission_count
            if self.support_raw_pdu():
                packet.metadata.raw = True
                if bytes.fromhex(packet.address[:2])[0] >= 0x80:
                    packet.preamble = 0xAA
                else:
                    packet.preamble = 0x55

            # Send packet
            return super().send_packet(packet)

        # Cannot send, return False.
        return False

    def support_raw_pdu(self):
        """
        Determine if the device supports raw PDU.
        """
        if self.__can_send_raw is None:
            capabilities = self.device.get_domain_capability(Domain.Esb)
            self.__can_send_raw = not capabilities & Capability.NoRawData
        return self.__can_send_raw


    def can_set_node_address(self):
        """
        Determine if the device can configure a Node address.
        """
        commands = self.device.get_domain_commands(Domain.Esb)
        return (
            (commands & (1 << Commands.SetNodeAddress)) > 0
        )

    def sniff(self, channel : int = None, address : str = "FF:FF:FF:FF:FF",
              show_acknowledgements : bool = False):
        """
        Sniff Enhanced ShockBurst packets.

        :param channel: Channel to listen, None to iterate over all possible channels
        :type channel: int
        :param address: Device address to target
        :type address: str
        :param show_acknowledgements: Sniff packets acknowledgements if set to True (default: False)
        :type show_acknowledgements: bool
        """
        if not self.can_sniff():
            raise UnsupportedCapability("Sniff")

        self.__cached_address = ESBAddress(address)
        if channel is None:
            # Enable scanning mode
            channel = 0xFF
        else:
            self.__cached_channel = channel

        msg = self.hub.esb.create_sniff_mode(
            EsbNodeAddress(self.__cached_address.value),
            channel,
            show_acknowledgements
        )
        resp = self.send_command(msg, message_filter(CommandResult))
        return isinstance(resp, Success)



    def can_be_prx(self):
        """
        Determine if the device implements a Primary Receiver (PRX) role mode.
        """
        commands = self.device.get_domain_commands(Domain.Esb)
        return (
            (commands & (1 << Commands.PrimaryReceiverMode)) > 0 and
            (commands & (1 << Commands.Start))>0 and
            (commands & (1 << Commands.Stop))>0
        )

    def enable_prx_mode(self, channel):
        """
        Enable Enhanced ShockBurst primary receiver (PRX) mode.
        """
        if not self.can_be_prx():
            raise UnsupportedCapability("PrimaryReceiverMode")

        # Check that we got a channel provided
        if channel is None:
            return False

        # Keep provided channel in cache
        self.__cached_channel = channel

        # Create a PrxMode message.
        msg = self.hub.esb.create_prx_mode(channel)

        resp = self.send_command(msg, message_filter(CommandResult))
        return isinstance(resp, Success)


    def can_be_ptx(self):
        """
        Determine if the device implements a Primary Transmitter (PTX) role mode.
        """
        commands = self.device.get_domain_commands(Domain.Esb)
        return (
            (commands & (1 << Commands.PrimaryTransmitterMode)) > 0 and
            (commands & (1 << Commands.Start))>0 and
            (commands & (1 << Commands.Stop))>0
        )

    def enable_ptx_mode(self, channel):
        """
        Enable Enhanced ShockBurst primary transmitter (PTX) mode.
        """
        if not self.can_be_ptx():
            raise UnsupportedCapability("PrimaryTransmitterMode")


        if channel is None:
            # Enable scanning mode
            channel = 0xFF
        else:
            # Keep channel in cache
            self.__cached_channel = channel

        # Create a PtxMode message.
        msg = self.hub.esb.create_ptx_mode(channel)

        resp = self.send_command(msg, message_filter(CommandResult))
        return isinstance(resp, Success)

    def set_node_address(self, address):
        """
        Enable Enhanced ShockBurst primary receiver (PRX) mode.
        """
        if not self.can_set_node_address():
            raise UnsupportedCapability("SetNodeAddress")

        node_address = ESBAddress(address)
        # Keep address in cache
        self.__cached_address = node_address

        # Create a SetNodeAddress message
        msg = self.hub.esb.create_set_node_address(
            EsbNodeAddress(node_address.value)
        )

        resp = self.send_command(msg, message_filter(CommandResult))
        return isinstance(resp, Success)

    def start(self):
        """
        Start currently enabled mode.
        """
        # Create a Start message.
        msg = self.hub.esb.create_start()

        resp = self.send_command(msg, message_filter(CommandResult))
        return isinstance(resp, Success)

    def stop(self):
        """
        Stop currently enabled mode.
        """
        # Create a Stop message.
        msg = self.hub.esb.create_stop()

        resp = self.send_command(msg, message_filter(CommandResult))
        return isinstance(resp, Success)

    def on_discovery_msg(self, message):
        pass

    def on_generic_msg(self, message):
        pass

    def on_domain_msg(self, domain, message):
        pass

    def on_packet(self, packet):
        """Incoming packet callback.
        """
        if not self.__ready:
            return

        # Dispatch packet
        if packet.metadata.raw:
            self.on_raw_pdu(packet)
        else:
            self.on_pdu(packet)

    def on_event(self, event):
        """Process incoming events.
        """
        if not self.__ready:
            return

        if isinstance(event, JammedEvt):
            self.on_jammed(event.timestamp)

    def on_raw_pdu(self, packet: Packet):
        """Callback to handle incoming raw PDU

        :param packet: Incoming PDU
        :type packet: Packet
        """
        # Extract the PDU from raw packet
        if ESB_Payload_Hdr in packet:
            pdu = packet[ESB_Payload_Hdr:]
        else:
            pdu = ESB_Payload_Hdr()/ESB_Ack_Response()

        # Propagate metadata to PDU
        pdu.metadata = packet.metadata
        self.on_pdu(pdu)

    def on_pdu(self, packet):
        """Callback for incoming PDU.

        For now, all compatible devices report raw PDUs when it comes to ESB,
        so this should not be called.
        """

    def on_jammed(self, timestamp: int):
        """Jammed event handler.
        """



"""
WHAD main device interface module.

This module provides multiple classes to handle WHAD device communication,
including background threads.
"""
import logging
from binascii import hexlify
from time import time, sleep
from queue import Queue, Empty
from threading import Thread, Lock

# Whad imports
from whad.exceptions import UnsupportedDomain, WhadDeviceNotReady, \
    WhadDeviceNotFound, WhadDeviceDisconnected, WhadDeviceTimeout, \
    WhadDeviceError
from whad.helpers import message_filter

# WHAD protocol hub
from whad.hub import ProtocolHub
from whad.hub.message import AbstractPacket, AbstractEvent
from whad.hub.generic.cmdresult import CommandResult
from whad.hub.discovery import InfoQueryResp, DomainInfoQueryResp, DeviceReady
from whad.hub.discovery import DeviceType
from whad.hub.generic.cmdresult import ResultCode

from whad.device.info import WhadDeviceInfo

logger = logging.getLogger(__name__)

class WhadDeviceInputThread(Thread):

    """WhadDevice I/O cancellable Thread.

    This thread runs in background and regularly calls the
    device read() method to fetch any incoming data.
    """

    def __init__(self, device, io_thread):
        super().__init__()
        self.__device = device
        self.__io_thread = io_thread
        self.__canceled = False
        self.daemon = True

    def cancel(self):
        """
        Cancel current I/O task.
        """
        self.__canceled = True

    def run(self):
        """
        Main task, call device process() method until thread
        is canceled.
        """
        while not self.__canceled:
            try:
                self.__device.read()
                #logger.debug("[WhadDeviceInputThread::run()] Read data from device")
            except WhadDeviceNotReady:
                logger.debug("Device %s has just disconnected (device not ready)",
                              self.__device.interface)
                break
            except WhadDeviceDisconnected:
                logger.debug("Device %s has just disconnected (read returned None)",
                             self.__device.interface)
                break
        logger.info("Device IO thread canceled and stopped.")
        self.__io_thread.on_disconnection()
        logger.info("on_disconnection done.")
        logger.info("Input thread for interface %s has stopped.", self.__device.interface)

class WhadDeviceMessageThread(Thread):
    """Main device incoming message processing thread.

    This thread runs in background, retrieves every message
    sent by the WHAD device to the host and forwards them to the
    `process_messages()` callback.
    """

    def __init__(self, device):
        super().__init__()
        self.__device = device
        self.__canceled = False
        self.daemon = True

    def cancel(self):
        """
        Cancel current I/O task.
        """
        self.__canceled = True

    def run(self):
        """
        Main task, call device process_messages() method until thread
        is canceled.
        """
        while not self.__canceled:
            self.__device.process_messages()

        # Finish processing remaining messages
        logger.debug("[WhadDeviceMessageThread] processing remaining messages ...")
        while self.__device.process_messages(timeout=.1):
            pass

        logger.info("Device message thread canceled and stopped, closing device %s.",
                    self.__device)
        self.__device.close()
        logger.info("Message processing thread for interface %s has stopped.", self.__device.interface)


class WhadDeviceIOThread:
    """Main device IO thread.

    This thread runs in background and is in charge of collecting
    incoming messages from a WHAD interface and sending pending messages
    to the WHAD interface.
    """

    def __init__(self, device):
        self.__device = device
        self.__input = WhadDeviceInputThread(device, self)
        self.__input.daemon = True
        self.__processing = WhadDeviceMessageThread(device)
        self.__processing.daemon = True
        self.__disconnected = False
        self.__alive = False

    def is_alive(self) -> bool:
        """Check if IO thread is still alive.
        """
        return self.__alive

    def cancel(self):
        """Cancel device IO management thread.
        """
        self.__input.cancel()
        self.__processing.cancel()

    def start(self):
        """Start device IO management thread.
        """
        logger.info("starting WhadDevice IO management thread ...")
        self.__input.start()
        self.__processing.start()
        logger.info("WhadDevice IO management thread up and running.")
        self.__alive = True

    def join(self):
        """Wait for device IO management thread to complete.
        """
        logger.info("waiting for WhadDevice IO management thread to finish ...")
        while not self.__disconnected:
            try:
                logger.info("Waiting for io thread ...")
                self.__input.join(1.0)
            except RuntimeError:
                logger.debug((
                    "RuntimeError raised while joining input thread, we may try"
                    " to wait for our thread to finish :/"))

            try:
                logger.info("Waiting for processing thread ...")
                self.__processing.join(1.0)
            except RuntimeError:
                logger.debug(("RuntimeError raised while joining processing thread,"
                              " we may try to wait for our thread to finish :/"))

        logger.info("WhadDevice IO management thread finished.")
        self.__alive = False
        if self.__device.opened:
            logger.info("Closing device due to IO termination.")
            self.__device.close()

    def on_disconnection(self):
        """Handle underlying device disconnection
        """
        logger.debug("[device::io_thread] Adapter has disconnected")
        self.__disconnected = True
        self.__input.cancel()
        #self.__device.close()
        self.__processing.cancel()


class WhadDevice:
    """
    WHAD Device interface class.

    This device class handles the device discovery process, every possible
    discovery and generic messages related to the device discovery. It MUST be
    inherited by device handling classes (such as the UartDevice class) in order
    to provide read/write capabilities.

    Inherited classes MUST only implement the following methods:

      * open(): will handle device opening/access
      * close(): will handle device closing
      * read() to read data from the device and send them to on_data_received()
      * write() to send data to the device

    All the message re-assembling, parsing, dispatching and background data reading
    will be performed in this class.
    """

    INTERFACE_NAME = None

    @classmethod
    def _get_sub_classes(cls):
        """
        Helper allowing to get every subclass of WhadDevice.
        """
        # List every available device class
        device_classes = set()
        for device_class in cls.__subclasses__():
            if device_class.__name__ == "VirtualDevice":
                for virtual_device_class in device_class.__subclasses__():
                    device_classes.add(virtual_device_class)
            else:
                device_classes.add(device_class)
        return device_classes

    @classmethod
    def create_inst(cls, interface_string):
        """
        Helper allowing to get a device according to the interface string provided.

        To make it work, every device class must implement:
            - a class attribute INTERFACE_NAME, matching the interface name
            - a class method list, returning the available devices
            - a property identifier, allowing to identify the device in a unique way

        This method should NOT be used outside of this class. Use WhadDevice.create instead.
        """

        if interface_string.startswith(cls.INTERFACE_NAME):
            identifier = None
            index = None
            if len(interface_string) == len(cls.INTERFACE_NAME):
                index = 0
            elif interface_string[len(cls.INTERFACE_NAME)] == ":":
                index = None
                try:
                    _, identifier = interface_string.split(":")
                except ValueError:
                    identifier = None
            else:
                try:
                    index = int(interface_string[len(cls.INTERFACE_NAME):])
                except ValueError:
                    index = None

            # Retrieve the list of available devices
            # (could be a list or a dict)
            available_devices = cls.list()

            # If the list of device is built statically, check before instantiation
            if available_devices is not None:
                if index is not None:
                    try:
                        # Try to retrieve a device based on the provided index
                        return available_devices[index]
                    except KeyError as exc:
                        raise WhadDeviceNotFound from exc
                    except IndexError as exc:
                        raise WhadDeviceNotFound from exc
                elif identifier is not None:
                    if isinstance(available_devices, list):
                        for dev in available_devices:
                            if dev.identifier == identifier:
                                return dev
                    elif isinstance(available_devices, dict):
                        for dev_id, dev in available_devices.items():
                            if dev.identifier == identifier:
                                return dev
                    raise WhadDeviceNotFound
                else:
                    raise WhadDeviceNotFound
            # Otherwise, check dynamically using check_interface
            else:
                formatted_interface_string = interface_string.replace(
                    cls.INTERFACE_NAME + ":",
                    ""
                )
                if cls.check_interface(formatted_interface_string):
                    return cls(formatted_interface_string)
                raise WhadDeviceNotFound

        else:
            raise WhadDeviceNotFound

    @classmethod
    def create(cls, interface_string):
        '''
        Create a specific device according to the provided interface string,
        formed as follows:

        <device_type>[device_index][:device_identifier]

        Examples:
            - `uart` or `uart0`: defines the first compatible UART device available
            - `uart1`: defines the second compatible UART device available
            - `uart:/dev/ttyACMO`: defines a compatible UART device identified by `/dev/tty/ACMO`
            - `ubertooth` or `ubertooth0`: defines the first available Ubertooth device
            - `ubertooth:11223344556677881122334455667788`: defines a Ubertooth device with serial number *11223344556677881122334455667788*
        '''
        device_classes = cls._get_sub_classes()

        device = None
        for device_class in device_classes:
            try:
                device = device_class.create_inst(interface_string)
                return device
            except WhadDeviceNotFound:
                continue

        raise WhadDeviceNotFound

    @classmethod
    def list(cls):
        '''
        Returns every available compatible devices.
        '''
        device_classes = cls._get_sub_classes()

        available_devices = []
        for device_class in device_classes:
            device_class_list = device_class.list()
            if device_class_list is not None:
                if isinstance(device_class_list, list):
                    for device in device_class_list:
                        available_devices.append(device)
                elif isinstance(device_class_list, dict):
                    for dev_id, device in device_class_list.items():
                        available_devices.append(device)
        return available_devices

    @classmethod
    def check_interface(cls, interface):
        '''
        Checks dynamically if the device can be instantiated.
        '''
        logger.debug("default: checking interface %s fails.", interface)
        return False

    @property
    def interface(self):
        '''
        Returns the current interface of the device.
        '''
        # If class has interface name, return the interface alias
        if hasattr(self.__class__,"INTERFACE_NAME"):
            return self.__class__.INTERFACE_NAME + str(self.index)

        # Interface is unknown
        return "unknown"

    @property
    def type(self):
        '''
        Returns the name of the class linked to the current device.
        '''
        return self.__class__.__name__


    @property
    def opened(self) -> bool:
        """Determine if the interface has already been opened

        :return: `True` if interface is already open, `False` otherwise
        :rtype: bool
        """
        return self.__opened

    def __init__(self, index: int = None):
        """Initialize a device

        Device index can be specified through the `index` argument, but keep
        in mind that this `index` argument, if used, must be passed for all
        devices of the same class in order not to mess up with numbering.
        Calling code will be in charge of keeping device indexes unique.

        :param index: Specifies the index of this device
        :type index: int
        """
        # Device information
        self.__info = None
        self.__discovered = False
        self.__opened = False
        self.__closing = False

        # Generate device index if not provided
        if index is None:
            self.inc_dev_index()
            self.__index = self.__class__.CURRENT_DEVICE_INDEX
        else:
            # Used by HCI devices to force index to match system names
            self.__index = index

        # Device connectors
        self.__connector = None

        # Device IO thread
        self.__io_thread = None

        # Default timeout for messages (5 seconds)
        self.__timeout = 5.0

        # Message queues
        self.__messages = Queue()
        self.__msg_queue = Queue()
        self.__mq_filter = None

        # Input pipes
        self.__inpipe = bytearray()

        # Create locks
        self.__lock = Lock()
        self.__tx_lock = Lock()

        # Protocol hub
        self.__hub = ProtocolHub(2)

    @property
    def hub(self):
        """Retrieve the device protocol hub (parser/factory)
        """
        return self.__hub

    @property
    def index(self) -> int:
        """Get the interface index

        :return: Interface index
        :rtype: int
        """
        return self.__index

    @classmethod
    def inc_dev_index(cls):
        """Inject and maintain device index.
        """
        if hasattr(cls, 'CURRENT_DEVICE_INDEX'):
            cls.CURRENT_DEVICE_INDEX += 1
        else:
            cls.CURRENT_DEVICE_INDEX = 0

    def lock(self):
        """Locks the pending output data buffer."""
        self.__lock.acquire()

    def unlock(self):
        """Unlocks the pending output data buffer."""
        self.__lock.release()

    def set_connector(self, connector):
        """
        Set this device connector.

        :param WhadDeviceConnector connector: connector to be used with this device.
        """
        self.__connector = connector

    ######################################
    # Device I/O operations
    ######################################

    def open(self):
        """
        Open device method. By default, creates a simple thread
        that will handle I/O in background. This requires the object
        to be ready for I/O operations when this method is called.

        This method MUST be overriden by inherited classes.
        """
        self.__io_thread = WhadDeviceIOThread(self)
        self.__io_thread.start()

        # Ask firmware for a reset
        try:
            logger.info("resetting device (if possible)")
            self.__opened = True
            self.reset()
        except Empty as err:
            # Device is unresponsive, does not seem compatible
            # Shutdown IO thread
            self.__io_thread.cancel()
            self.__io_thread.join()

            # Notify device not found
            raise WhadDeviceNotReady() from err

    def close(self):
        """
        Close device.

        This method MUST be overriden by inherited classes.
        """
        # Avoid recursion when closing
        if not self.__opened or self.__closing:
            logger.debug("exiting close() to avoid recursion")
            return

        logger.info("closing WHAD device")
        self.__closing = True

        # Cancel I/O thread if required
        if self.__io_thread is not None:
            if self.__io_thread.is_alive():
                self.__io_thread.cancel()

        # Send a NOP message to unlock process_messages()
        logger.debug("send NOP message")
        msg = self.hub.generic.create_verbose(b'')
        self.on_message_received(msg)
        logger.debug("NOP message sent")

        # Wait for the thread to terminate nicely.
        if self.__io_thread is not None:
            self.__io_thread.join()

        self.__opened = False
        self.__closing = False

        # Notify connector device has closed
        if self.__connector is not None:
            self.__connector.on_disconnection()

    def wait(self):
        """Wait for device IO management thread to stop.
        """
        logger.debug("[WhadDevice] waiting for background thread to gracefully stop")
        self.__io_thread.join()
        logger.debug("[WhadDevice] background threads stopped")

    def is_open(self):
        """Determine if the device has been opened or not.
        """
        return self.__opened

    def __write(self, data):
        """
        Sends data to the device.

        This is an internal method that SHALL NOT be used from inherited classes.
        """
        self.lock()
        logger.debug("sending %s to WHAD device %s", bytes(data), self.interface)
        self.write(bytes(data))
        self.unlock()

    def write(self, data: bytes) -> int:
        """Default write method. This implementation emulates a successful write,
        but classes inheriting from WhadDevice must subclass this method to provide
        their own implementation.

        :param data: Data to write to the WHAD interface
        :type data: bytes
        :return: Number of bytes effectively written to the interface
        :rtype: int
        """
        return len(data)

    def set_queue_filter(self, queue_filter=None):
        """Sets the message queue filter.

        :param filter: filtering function/lambda to be used by our message queue filter.
        """
        logger.debug("set queue filter: %s", queue_filter)
        self.__mq_filter = queue_filter

    def wait_for_single_message(self, timeout, msg_filter=None):
        """Configures the device message queue filter to automatically move messages
        that matches the filter into the queue, and then waits for the first message
        that matches this filter and returns it.
        """
        if filter is not None:
            self.set_queue_filter(msg_filter)

        # Wait for a matching message to be caught (blocking)
        return self.__msg_queue.get(block=True, timeout=timeout)


    def wait_for_message(self, timeout=None, filter=None, command=False):
        """
        Configures the device message queue filter to automatically move messages
        that matches the filter into the queue, and then waits for the first message
        that matches this filter and process it.

        This method is blocking until a matching message is received.

        :param int timeout: Timeout
        :param filter: Message queue filtering function (optional)
        """

        # Check if device is still opem
        if not self.__opened and self.__msg_queue.empty():
            raise WhadDeviceDisconnected()

        logger.debug("entering wait_for_message ...")
        if filter is not None:
            self.set_queue_filter(filter)

        start_time = time()

        while True:
            try:
                # Wait for a matching message to be caught (blocking)
                msg = self.__msg_queue.get(block=True, timeout=timeout)

                # If message does not match, dispatch.
                if not self.__mq_filter(msg):
                    self.dispatch_message(msg)
                    logger.debug("exiting wait_for_message ...")
                else:
                    logger.debug("exiting wait_for_message ...")
                    return msg
            except Empty as err:
                # Queue is empty, wait for a message to show up.
                if timeout is not None and (time() - start_time > timeout):
                    if command:
                        raise WhadDeviceTimeout("WHAD device did not answer to a command") from err

                    logger.debug("exiting wait_for_message ...")
                    return None

                sleep(0.001)



    def send_message(self, message, keep=None):
        """
        Serializes a message and sends it to the device, without waiting for an answer.
        Optionally, you can update the message queue filter if you need to wait for
        specific messages after the message sent.

        :param Message message: Message to send
        :param keep: Message queue filter function
        """
        logger.info("sending message %s to device <%s>", message, self.interface)

        # if `keep` is set, configure queue filter
        if keep is not None:
            logger.debug("send_message:set_queue_filter")
            self.set_queue_filter(keep)

        # Convert message into bytes
        raw_message = message.serialize()


        # Define header
        header = [
            0xAC, 0xBE,
            len(raw_message) & 0xff,
            (len(raw_message) >> 8) & 0xff
        ]

        # Send header followed by serialized message
        self.__write(header)
        self.__write(raw_message)


    def send_command(self, command, keep=None):
        """
        Sends a command and awaits a specific response from the device.
        WHAD commands usualy expect a CmdResult message, if `keep` is not
        provided then this method will by default wait for a CmdResult.

        :param Message command: Command message to send to the device
        :param keep: Message queue filter function (optional)
        :returns: Response message from the device
        :rtype: Message
        """
        with self.__tx_lock:

            # If a queue filter is not provided, expect a default CmdResult
            try:
                if keep is None:
                    self.send_message(command, message_filter(CommandResult))
                else:
                    self.send_message(command, keep)
            except WhadDeviceError as error:
                # Device error has been triggered, it looks like our device is in
                # an unspecified state, notify user.
                logger.debug("WHAD device in error while sending message: %s", error)
                raise error

            try:
                # Retrieve the first message matching our filter.
                result = self.wait_for_message(self.__timeout, command=True)
            except WhadDeviceTimeout as timedout:
                # Forward exception
                raise timedout

        # Log message
        logger.debug("Command result: %s", result)

        return result


    def on_data_received(self, data):
        """
        Data received callback.

        This callback will process incoming messages, parse them
        and then forward to the message processing callback.

        :param bytes data: Data received from the device.
        """
        #logger.info("[WhadDevice] entering on_data_received()")
        logger.debug("[WhadDevice] received raw data from device <%s>: %s",
                     self.interface, hexlify(data))
        self.__inpipe.extend(data)
        while len(self.__inpipe) > 2:
            # Is the magic correct ?
            if self.__inpipe[0] == 0xAC and self.__inpipe[1] == 0xBE:
                # Have we received a complete message ?
                if len(self.__inpipe) > 4:
                    msg_size = self.__inpipe[2] | (self.__inpipe[3] << 8)
                    if len(self.__inpipe) >= (msg_size+4):
                        raw_message = self.__inpipe[4:4+msg_size]

                        # Parse received message with our Protocol Hub
                        msg = self.__hub.parse(bytes(raw_message))

                        # Forward message if successfully parsed
                        if msg is not None:
                            self.on_message_received(msg)

                        # Chomp
                        self.__inpipe = self.__inpipe[msg_size + 4:]
                    else:
                        break
                else:
                    break
            else:
                # Nope, that's not a header
                while len(self.__inpipe) >= 2:
                    if (self.__inpipe[0] != 0xAC) or (self.__inpipe[1] != 0xBE):
                        self.__inpipe = self.__inpipe[1:]
                    else:
                        break


    def dispatch_message(self, message):
        """Dispatches an incoming message to the corresponding callbacks depending on its
        type and content.

        :param Message message: Message to dispatch
        """
        logger.info("dispatching WHAD message ...")

        # Allows a connector to catch any message
        self.on_any_msg(message)

        # Forward to dedicated callbacks
        if message.message_type == "discovery":
            logger.info("message is about device discovery, forwarding to discovery handler")
            self.on_discovery_msg(message)
        elif message.message_type == "generic":
            logger.info("message is generic, forwarding to default handler")
            self.on_generic_msg(message)
            logger.info("on_generic_message called")
        else:
            domain = message.message_type
            if domain is not None:
                logger.info("message concerns domain `%s`, forward to domain-specific handler",
                            domain)
                self.on_domain_msg(domain, message)

    def on_message_received(self, message):
        """
        Method called when a WHAD message is received, dispatching.

        :param Message message: Message received
        """
        if self.__closing:
            return

        logger.debug(("[WhadDevice::on_message_received()][%s] "
                      "message received: %s"), self.interface, message)
        logger.debug(("[WhadDevice::on_message_received()][%s] "
                     "message queue filter: %s"), self.interface, self.__mq_filter)

        # If message queue filter is defined and message matches this filter,
        # move it into our message queue.
        if self.__mq_filter is not None and self.__mq_filter(message):
            logger.info("message does match current filter, save it for processing")
            self.__msg_queue.put(message, block=True)
            logger.info("message added to message queue")
        else:
            # Save message for background dispatch
            logger.info(("message does not match filter or no filter set, "
                         "save in default message queue"))
            self.__messages.put(message, block=True)
            logger.info("message added to default message queue")

    def has_pending_messages(self) -> bool:
        """Determine if the device has sent some messages.

        :return: True if there are some messages awaiting for processing, False otherwise
        :rtype: bool
        """
        return not self.__messages.empty()

    def process_messages(self, timeout=1.0) -> bool:
        """Process pending messages
        """
        result = False

        #if self.__closing:
        #    return

        # If no connector set, we cannot process messages, we keep them in
        # memory instead.
        if self.__connector is None:
            return False

        try:
            message = self.__messages.get(block=True, timeout=timeout)
            if message is not None:
                logger.debug("[process_messages] retrieved message %s", message)
                self.dispatch_message(message)
                result = True

            return result
        except Empty:
            return False

    ######################################
    # Any messages
    ######################################

    def on_any_msg(self, message):
        """This callback method is called when any message is received

        :param Message message: WHAD message received
        """
        logger.debug("on_any_msg")
        # Forward message to the connector, if any
        if self.__connector is not None:
            logger.debug("Forward message %s to connector %s",
                         message, self.__connector)
            self.__connector.on_any_msg(message)

    ######################################
    # Generic messages handling
    ######################################

    def on_generic_msg(self, message):
        """
        This callback method is called whenever a Generic message is received.

        :param Message message: Generic message received
        """
        # Handle generic result message
        if isinstance(message, CommandResult):
            if message.result_code == CommandResult.UNSUPPORTED_DOMAIN:
                logger.error("domain not supported by this device")
                raise UnsupportedDomain("")

        # Forward everything to the connector, if any
        if self.__connector is not None:
            logger.debug("Forward generic message %s to connector %s",
                         message, self.__connector)
            self.__connector.on_generic_msg(message)
        else:
            logger.debug("No connector registered, message lost")


    ######################################
    # Generic discovery
    ######################################

    def on_discovery_msg(self, message):
        """
        Method called when a discovery message is received. If a connector has
        been associated with the device, forward this message to this connector.
        """

        # Forward everything to the connector, if any
        if self.__connector is not None:
            self.__connector.on_discovery_msg(message)

    def has_domain(self, domain) -> bool:
        """Checks if device supports a specific domain.

        :param Domain domain: Domain
        :returns: True if domain is supported, False otherwise.
        :rtype: bool
        """
        if self.__info is not None:
            return self.__info.has_domain(domain)

        # No info available on device, domain is not supported by default
        return False


    def get_domains(self) -> dict:
        """Get device' supported domains.

        :returns: list of supported domains
        :rtype: list
        """
        if self.__info is not None:
            return self.__info.domains

        # No domain discovered yet
        return {}


    def get_domain_capability(self, domain):
        """Get a device domain capabilities.

        :param Domain domain: Target domain
        :returns: Domain capabilities
        :rtype: DeviceDomainInfoResp
        """
        if self.__info is not None:
            return self.__info.get_domain_capabilities(domain)

        # No capability if not discovered
        return 0

    def get_domain_commands(self, domain):
        """Get a device supported domain commands.

        :param Domain domain: Target domain
        :returns: Bitmask of supported commands
        :rtype: int
        """
        if self.__info is not None:
            return self.__info.get_domain_commands(domain)

        # No supported commands by default
        return 0


    def send_discover_info_query(self, proto_version=0x0100):
        """
        Sends a DeviceInfoQuery message and awaits for a DeviceInfoResp
        answer.
        """
        logger.info("preparing a DeviceInfoQuery message")
        msg = self.__hub.discovery.create_info_query(proto_version)
        return self.send_command(
            msg,
            message_filter(InfoQueryResp)
        )


    def send_discover_domain_query(self, domain):
        """
        Sends a DeviceDomainQuery message and awaits for a DeviceDomainResp
        answer.
        """
        logger.info("preparing a DeviceDomainInfoQuery message")
        msg = self.__hub.discovery.create_domain_query(domain)
        return self.send_command(
            msg,
            message_filter(DomainInfoQueryResp)
        )

    def discover(self):
        """
        Performs device discovery (synchronously).

        Discovery process asks the device to provide its description, including
        its supported domains and associated capabilities. For each domain we
        then query the device and get the list of supported commands.
        """
        if not self.__discovered:
            # We send a DeviceInfoQuery message to the device and expect a
            # DeviceInfoResponse in return.
            resp = self.send_discover_info_query()

            # If we have an answer, process it.
            if resp is not None:

                # Ensure response is the one we expect
                assert isinstance(resp, InfoQueryResp)

                # Save device information
                self.__info = WhadDeviceInfo(
                    resp
                )

                # Parse DeviceInfoResponse
                #device_info = self.hub.parse(resp)

                # Update our ProtocolHub version to the device version
                self.__hub = ProtocolHub(resp.proto_min_ver)

                # Query device domains
                logger.info("query supported commands per domain")
                for domain in self.__info.domains:
                    resp = self.send_discover_domain_query(domain)
                    self.__info.add_supported_commands(
                        resp.domain,
                        resp.supported_commands
                    )

                # Mark device as discovered
                logger.info("device discovery done")
                self.__discovered = True

                # Switch to max transport speed
                logger.info("set transport speed to %d", self.info.max_speed)
                self.change_transport_speed(
                    self.info.max_speed
                )
            else:
                logger.error("device is not ready !")
                raise WhadDeviceNotReady()

    def reset(self):
        """Reset device
        """
        logger.info("preparing a DeviceResetQuery message")
        msg = self.hub.discovery.create_reset_query()
        return self.send_command(
            msg,
            message_filter(DeviceReady)
        )

    def change_transport_speed(self, speed):
        """Set device transport speed.

        Optional.
        """

    @property
    def device_id(self):
        """Return device ID
        """
        return self.__info.device_id

    @property
    def info(self) -> WhadDeviceInfo:
        """Get device info object

        :return: Device information object
        :rtype: WhadDeviceInfo
        """
        return self.__info

    ######################################
    # Upper layers (domains) handling
    ######################################

    def on_domain_msg(self, domain, message):
        """
        Callback method handling domain-related messages. Since this layer is not
        managed by the root WhadDevice class, forward it to the upper layer, the
        associated connector (if any).

        :param Domain domain: Target domain
        :param Message message: Domain-related message received
        """

        # Forward everything to the connector, if any
        if self.__connector is not None:

            # Check if message is a received packet
            if issubclass(message, AbstractPacket):
                # If connector is locked, save message into locked pdus
                if self.__connector.is_locked():
                    self.__connector.add_locked_pdu(message)
                else:
                    self.on_packet_message(message)
            elif issubclass(message, AbstractEvent):
                # Convert message into event
                event = message.to_event()
                if event is not None:
                    # Forward to our connector
                    self.__connector.on_event(event)
            else:
                # Forward other messages to on_domain_msg() callback
                self.__connector.on_domain_msg(domain, message)
        return False

    def on_packet_message(self, message):
        """Process packet message
        """
        # Convert message into packet
        packet = message.to_packet()
        if packet is not None:

            # Report packet to monitors
            self.__connector.monitor_packet_rx(packet)

            # Forward to our connector
            if self.__connector.is_synchronous():
                # If we are in synchronous mode, add packet as pending
                self.__connector.add_pending_packet(packet)
            else:
                #logger.debug("[WhadDevice] on_domain_msg() for device %s: %s",
                #             self.interface, message)
                self.__connector.on_packet(packet)

class VirtualDevice(WhadDevice):
    """
    AdapterDevice device class.
    """
    def __init__(self, index: int = None):
        self._dev_type = None
        self._dev_id = None
        self._fw_author = None
        self._fw_url = None
        self._fw_version = (0, 0, 0)
        self._dev_capabilities = {}
        self.__lock = Lock()
        super().__init__(index)

    def send_message(self, message, keep=None):
        """Send message to host.
        """
        with self.__lock:
            super().set_queue_filter(keep)
            self._on_whad_message(message)

    def _on_whad_message(self, message):
        """TODO: associate callbacks with classes ?
        """
        category = message.message_type
        message_type = message.message_name

        callback_name = f"_on_whad_{category}_{message_type}"
        if hasattr(self, callback_name) and callable(getattr(self, callback_name)):
            getattr(self, callback_name)(message)
        else:
            logger.info("unhandled message: %s", message)
            self._send_whad_command_result(ResultCode.ERROR)

    def _on_whad_discovery_info_query(self, message):
        major, minor, revision = self._fw_version
        msg = self.hub.discovery.create_info_resp(
            DeviceType.VirtualDevice,
            self._dev_id,
            0x0100,
            0,
            self._fw_author,
            self._fw_url,
            major, minor, revision,
            [domain | (capabilities[0] & 0xFFFFFF) for domain, capabilities in self._dev_capabilities.items()]
        )
        self._send_whad_message(msg)

    def _on_whad_discovery_domain_query(self, message):
        # Compute supported commands for domain
        commands = 0
        supported_commands = self._dev_capabilities[message.domain][1]
        for command in supported_commands:
            commands |= (1 << command)

        # Create a DomainResp message and send it
        msg = self.hub.discovery.create_domain_resp(
            message.domain,
            commands
        )
        self._send_whad_message(msg)


    def _send_whad_message(self, message):
        self.on_message_received(message)

    def _send_whad_command_result(self, code):
        msg = self.hub.generic.create_command_result(code)
        self._send_whad_message(msg)



"""
WHAD Device module

This module provides a set of classes used to interact with devices
running compatible firmwares, as well as the default connector class used
to handle messages coming from or sent to the device.
"""
# Logging
import warnings

from cryptography.utils import CryptographyDeprecationWarning

# Whad device-related classes
from whad.device.info import WhadDeviceInfo
from whad.device.connector import WhadDeviceConnector
from whad.device.bridge import Bridge
from whad.device.device import WhadDevice, VirtualDevice

# Import derived classes
from whad.device.uart import UartDevice
from whad.device.tcp import TCPSocketDevice
from whad.device.unix import UnixSocketDevice
from whad.device.virtual import HCIDevice, APIMoteDevice, RFStormDevice, RZUSBStickDevice, \
    UbertoothDevice, YardStickOneDevice

# Remove scapy deprecation warnings
warnings.filterwarnings("ignore", category=CryptographyDeprecationWarning)

__all__ = [
    "Bridge",
    "WhadDeviceConnector",
    "WhadDeviceInfo",
    "WhadDevice",
    "UartDevice",
    "VirtualDevice",
    "TCPSocketDevice",
    "UnixSocketDevice",
    "HCIDevice",
    "APIMoteDevice",
    "RFStormDevice",
    "RZUSBStickDevice",
    "UbertoothDevice",
    "YardStickOneDevice"
]



"""WHAD new hardware interface
"""
import logging
import contextlib
from time import time, sleep
from typing import Generator, Callable
from threading import Thread, Lock
from queue import Queue, Empty

from whad.exceptions import WhadDeviceNotReady, WhadDeviceError, \
    WhadDeviceTimeout, WhadDeviceDisconnected, WhadDeviceNotFound
from whad.hub import ProtocolHub
from whad.hub.message import HubMessage
from whad.helpers import message_filter
from whad.hub.generic.cmdresult import CommandResult, ResultCode
from whad.hub.discovery import InfoQueryResp, DomainInfoQueryResp, DeviceReady
from whad.hub.discovery import DeviceType
from whad.device.info import WhadDeviceInfo

logger = logging.getLogger(__name__)

class IfaceEvt:
    """Interface event.
    """
    def __init__(self, iface = None):
        self.__iface = iface

    def __repr__(self) -> str:
        """Printable representation of this event
        """
        return f"IfaceEvt(iface='{self.__iface.name}')"

class Disconnected(IfaceEvt):
    """Interface has disconnected.
    """
    def __repr__(self) -> str:
        """Printable representation of this event.
        """
        return f"DisconnectedEvt(iface='{self.__iface.name}')"

class MessageReceived(IfaceEvt):
    """Message has been received.
    """
    def __init__(self, iface = None, message = None):
        super().__init__(iface)
        self.__message = message

    @property
    def message(self) -> HubMessage:
        """Received message."""
        return self.__message

    def __repr__(self) -> str:
        """Printable representation of this event.
        """
        return f"WhadMessageReceived(iface='{self.__iface.name}')"

class IfaceInThread(Thread):
    """Internal thread processing data sent by the connector to the
    hardware interface.
    """

    def __init__(self, iface = None):
        super().__init__()
        self.__iface = iface
        self.__canceled = False

    def cancel(self):
        """Cancel thread
        """
        self.__canceled = True

    def serialize(self, message) -> bytes:
        """Serialize a WHAD message.
        """
        # Serialize protobuf message
        raw_msg = message.serialize()

        # Define header
        header = [
            0xAC, 0xBE,
            len(raw_msg) & 0xff,
            (len(raw_msg) >> 8) & 0xff
        ]

        # Build the final payload
        return bytes(header) + raw_msg

    def run(self):
        """Out thread main task.
        """
        while not self.__canceled:
            # Read data from device (may block)
            try:
                # Wait for a message to send to interface (blocking)
                logger.debug("[in_thread] waiting for message to send")
                with self.__iface.get_pending_message(timeout=1.0) as message:
                    logger.debug("[in_thread] sending message %s", message)
                    # Serialize message and send it.
                    payload = self.serialize(message)

                    # Send serialized message to interface
                    logger.debug("[in_thread] acquiring lock on interface ...")
                    self.__iface.lock()
                    logger.debug("[in_thread] sending payload %s ...", payload)
                    self.__iface.write(payload)
                    logger.debug("[in_thread] releasing lock ...")
                    self.__iface.unlock()

                    # Notify message has correctly been sent, from a dedicated
                    # thread.
                    if message.has_callback():
                        Thread(target=message.sent).start()
            except Empty:
                pass
            except WhadDeviceNotReady:
                if message.has_callback():
                    Thread(target=message.error, args=[1]).start()
                break
            except WhadDeviceDisconnected:
                if message.has_callback():
                    Thread(target=message.error, args=[2]).start()
                break

class IfaceOutThread(Thread):
    """Internal thread processing data sent by the hardware interface
    to the interface object.
    """
    def __init__(self, iface = None):
        super().__init__()
        self.__iface = iface
        self.__canceled = False

        # Data processing
        self.__data = bytearray()

    def cancel(self):
        """Cancel thread
        """
        self.__canceled = True

    def ingest(self, data: bytes):
        """Ingest incoming bytes.
        """
        self.__data.extend(data)
        while len(self.__data) > 2:
            # Is the magic correct ?
            if self.__data[0] == 0xAC and self.__data[1] == 0xBE:
                # Have we received a complete message ?
                if len(self.__data) > 4:
                    msg_size = self.__data[2] | (self.__data[3] << 8)
                    if len(self.__data) >= (msg_size+4):
                        raw_message = self.__data[4:4+msg_size]

                        # Parse received message with our Protocol Hub
                        msg = self.__iface.hub.parse(bytes(raw_message))

                        # Forward message if successfully parsed
                        if msg is not None:
                            self.__iface.put_message(msg)

                        # Chomp
                        self.__data = self.__data[msg_size + 4:]
                    else:
                        break
                else:
                    break
            else:
                # Nope, that's not a header
                while len(self.__data) >= 2:
                    if (self.__data[0] != 0xAC) or (self.__data[1] != 0xBE):
                        self.__data = self.__data[1:]
                    else:
                        break

    def run(self):
        """Out thread main task.
        """
        while not self.__canceled:
            # Read data from device (may block)
            try:
                data = self.__iface.read()
                if data is not None:
                    self.ingest(data)
            except WhadDeviceNotReady:
                break
            except WhadDeviceDisconnected:
                break

class Interface:
    """WHAD 'new' hardware interface.
    """
    INTERFACE_NAME = None

    @classmethod
    def _get_sub_classes(cls):
        """
        Helper allowing to get every subclass of WhadDevice.
        """
        # List every available device class
        device_classes = set()
        for device_class in cls.__subclasses__():
            if device_class.__name__ == "VirtualDevice":
                for virtual_device_class in device_class.__subclasses__():
                    device_classes.add(virtual_device_class)
            else:
                device_classes.add(device_class)
        return device_classes

    @classmethod
    def create_inst(cls, interface_string):
        """
        Helper allowing to get a device according to the interface string provided.

        To make it work, every device class must implement:
            - a class attribute INTERFACE_NAME, matching the interface name
            - a class method list, returning the available devices
            - a property identifier, allowing to identify the device in a unique way

        This method should NOT be used outside of this class. Use WhadDevice.create instead.
        """

        if interface_string.startswith(cls.INTERFACE_NAME):
            identifier = None
            index = None
            if len(interface_string) == len(cls.INTERFACE_NAME):
                index = 0
            elif interface_string[len(cls.INTERFACE_NAME)] == ":":
                index = None
                try:
                    _, identifier = interface_string.split(":")
                except ValueError:
                    identifier = None
            else:
                try:
                    index = int(interface_string[len(cls.INTERFACE_NAME):])
                except ValueError:
                    index = None

            # Retrieve the list of available devices
            # (could be a list or a dict)
            available_devices = cls.list()

            # If the list of device is built statically, check before instantiation
            if available_devices is not None:
                if index is not None:
                    try:
                        # Try to retrieve a device based on the provided index
                        return available_devices[index]
                    except KeyError as exc:
                        raise WhadDeviceNotFound from exc
                    except IndexError as exc:
                        raise WhadDeviceNotFound from exc
                elif identifier is not None:
                    if isinstance(available_devices, list):
                        for dev in available_devices:
                            if dev.identifier == identifier:
                                return dev
                    elif isinstance(available_devices, dict):
                        for dev_id, dev in available_devices.items():
                            if dev.identifier == identifier:
                                return dev
                    raise WhadDeviceNotFound
                else:
                    raise WhadDeviceNotFound
            # Otherwise, check dynamically using check_interface
            else:
                formatted_interface_string = interface_string.replace(
                    cls.INTERFACE_NAME + ":",
                    ""
                )
                if cls.check_interface(formatted_interface_string):
                    return cls(formatted_interface_string)
                raise WhadDeviceNotFound

        else:
            raise WhadDeviceNotFound

    @classmethod
    def create(cls, interface_string):
        '''
        Create a specific device according to the provided interface string,
        formed as follows:

        <device_type>[device_index][:device_identifier]

        Examples:
            - `uart` or `uart0`: defines the first compatible UART device available
            - `uart1`: defines the second compatible UART device available
            - `uart:/dev/ttyACMO`: defines a compatible UART device identified by `/dev/tty/ACMO`
            - `ubertooth` or `ubertooth0`: defines the first available Ubertooth device
            - `ubertooth:11223344556677881122334455667788`: defines a Ubertooth device with serial number *11223344556677881122334455667788*
        '''
        device_classes = cls._get_sub_classes()

        device = None
        for device_class in device_classes:
            try:
                device = device_class.create_inst(interface_string)
                return device
            except WhadDeviceNotFound:
                continue

        raise WhadDeviceNotFound

    @classmethod
    def list(cls):
        '''
        Returns every available compatible devices.
        '''
        device_classes = cls._get_sub_classes()

        available_devices = []
        for device_class in device_classes:
            device_class_list = device_class.list()
            if device_class_list is not None:
                if isinstance(device_class_list, list):
                    for device in device_class_list:
                        available_devices.append(device)
                elif isinstance(device_class_list, dict):
                    for dev_id, device in device_class_list.items():
                        available_devices.append(device)
        return available_devices

    @classmethod
    def check_interface(cls, interface):
        '''
        Checks dynamically if the device can be instantiated.
        '''
        logger.debug("default: checking interface %s fails.", interface)
        return False

    @property
    def interface(self):
        '''
        Returns the current interface of the device.
        '''
        # If class has interface name, return the interface alias
        if hasattr(self.__class__,"INTERFACE_NAME"):
            return self.__class__.INTERFACE_NAME + str(self.index)

        # Interface is unknown
        return "unknown"

    @property
    def type(self):
        '''
        Returns the name of the class linked to the current device.
        '''
        return self.__class__.__name__

    def __init__(self, index: int = None):
        """Initialize an interface
        """
        # Interface state
        self.__info = None
        self.__opened = False
        self.__discovered = False
        self.__closing = False

        # Generate device index if not provided
        if index is None:
            self.inc_dev_index()
            self.__index = self.__class__.CURRENT_DEVICE_INDEX
        else:
            # Used by HCI devices to force index to match system names
            self.__index = index

        # IO Threads
        self.__iface_in = None
        self.__iface_out = None

        # Queue holding messages from connector, waiting to be sent to
        # the interface.
        self.__in_messages = Queue()

        # Queue holding messages from interface, waiting to be sent to
        # an attached connector
        self.__out_messages = Queue()

        # Connector bound to this device
        self.__connector = None

        # Interface lock
        self.__lock = Lock()

        # Connector lock
        self.__msg_filter: Callable[..., bool] = None

        # Protocol hub
        self.__hub = ProtocolHub()

        # Communication timeout
        self.__timeout = 5.0

    @contextlib.contextmanager
    def get_pending_message(self, timeout: float = None) -> Generator[HubMessage, None, None]:
        """Get message waiting to be sent to the interface.
        """
        try:
            yield self.__in_messages.get(timeout=timeout)
        except Empty as err:
            raise err
        else:
            self.__in_messages.task_done()

    @property
    def connector(self):
        """Connector bound to the interface
        """
        return self.__connector

    @property
    def hub(self):
        """Retrieve the device protocol hub (parser/factory)
        """
        return self.__hub

    @property
    def index(self) -> int:
        """Get the interface index

        :return: Interface index
        :rtype: int
        """
        return self.__index

    @property
    def device_id(self):
        """Return device ID
        """
        return self.__info.device_id

    @property
    def info(self) -> WhadDeviceInfo:
        """Get device info object

        :return: Device information object
        :rtype: WhadDeviceInfo
        """
        return self.__info

    @property
    def opened(self) -> bool:
        """Device is open ?
        """
        return self.is_open()

    def is_open(self) -> bool:
        """Determine if interface is opened.
        """
        return self.__opened

    @classmethod
    def inc_dev_index(cls):
        """Inject and maintain device index.
        """
        if hasattr(cls, 'CURRENT_DEVICE_INDEX'):
            cls.CURRENT_DEVICE_INDEX += 1
        else:
            cls.CURRENT_DEVICE_INDEX = 0

    def set_connector(self, connector):
        """Set interface connector.
        """
        self.__connector = connector

    def lock(self):
        """Lock interface for read/write operation.
        """
        self.__lock.acquire()
        logger.debug("Lock acquired !")

    def unlock(self):
        """Unlock interface for read/write operation.
        """
        logger.debug("Releasing lock ...")
        self.__lock.release()

    def __start_io_threads(self):
        """Start background IO threads
        """
        self.__iface_in = IfaceInThread(self)
        self.__iface_in.start()
        self.__iface_out= IfaceOutThread(self)
        self.__iface_out.start()

    def __stop_io_threads(self):
        """Stop background IO threads
        """
        if self.__iface_in is not None:
            self.__iface_in.cancel()
            self.__iface_in.join()
        if self.__iface_out is not None:
            self.__iface_out.cancel()
            self.__iface_out.join()

    ##
    # Device specific methods
    ##

    def open(self):
        """Handle device open
        """
        # Create interface I/O threads and start them.
        self.__start_io_threads()

        # Ask interface for a reset
        try:
            logger.info("resetting interface (if possible)")
            self.__opened = True
            self.reset()
        except Empty as err:
            # Device is unresponsive, shutdown IO threads
            self.__stop_io_threads()

            raise WhadDeviceNotReady() from err

    def read(self) -> bytes:
        """Read bytes from interface (blocking).
        """
        return b''

    def write(self, payload: bytes) -> int:
        """Write payload to interface.
        """
        return len(payload)

    def close(self):
        """Close device
        """
        logger.info("closing WHAD interface")
        self.__closing = True

        # Cancel I/O thread if required
        self.__stop_io_threads()

        self.__opened = False
        self.__closing = False

        # Notify connector that device has closed
        if self.__connector is not None:
            self.__connector.send_event(Disconnected(self))

    def change_transport_speed(self, speed):
        """Set device transport speed.

        Optional.
        """

    ##
    # Message processing
    ##

    def set_queue_filter(self, keep: Callable[..., bool] = None):
        """Set message queue filter.
        """
        self.__msg_filter = keep

    def put_message(self, message: HubMessage):
        """Process incoming message.
        """
        # If no connector is attached to the interface, redirect to a dedicated
        # message queue.
        if self.__connector is None:
            self.__out_messages.put(message)

        # If a connector is attached to the interface but a message filter
        # is set, redirect matching messages to a dedicated message queue
        # and notify the connector about the other messages.
        elif self.__msg_filter is not None and self.__msg_filter(message):
            self.__out_messages.put(message)
        else:
            self.connector.send_event(MessageReceived(self, message))


    def wait_for_single_message(self, timeout: float = None ,
                                msg_filter: Callable[..., bool] = None):
        """Configures the device message queue filter to automatically move messages
        that matches the filter into the queue, and then waits for the first message
        that matches this filter and returns it.
        """
        unexpected_messages = []
        if msg_filter is not None:
            self.set_queue_filter(msg_filter)

        # Wait for a matching message to be caught (blocking)
        msg = self.__out_messages.get(block=True, timeout=timeout)

        # If message filter is set and message does not match, wait until an
        # expected message matches
        if msg_filter is not None:
            # Wait for a matching message
            while not msg_filter(msg):
                unexpected_messages.append(msg)
                msg = self.__out_messages.get(block=True, timeout=timeout)

            # Re-enqueue non-matching messages
            for m in unexpected_messages:
                self.__out_messages.put(m)
        return msg


    def wait_for_message(self, timeout: float = None, msg_filter: Callable[..., bool] = None,
                         command: bool = False):
        """
        Configures the device message queue filter to automatically move messages
        that matches the filter into the queue, and then waits for the first message
        that matches this filter and process it.

        This method is blocking until a matching message is received.

        :param int timeout: Timeout
        :param filter: Message queue filtering function (optional)
        """

        # Check if device is still opem
        if not self.__opened and self.__out_messages.empty():
            raise WhadDeviceDisconnected()

        logger.debug("entering wait_for_message ...")
        if msg_filter is not None:
            self.set_queue_filter(msg_filter)

        start_time = time()

        while True:
            try:
                # Wait for a matching message to be caught (blocking)
                msg = self.__out_messages.get(block=True, timeout=timeout)

                # If message does not match, re-enqueue
                if not self.__msg_filter(msg):
                    self.put_message(msg)
                    logger.debug("exiting wait_for_message ...")
                else:
                    logger.debug("exiting wait_for_message ...")
                    return msg
            except Empty as err:
                # Queue is empty, wait for a message to show up.
                if timeout is not None and (time() - start_time > timeout):
                    if command:
                        raise WhadDeviceTimeout("WHAD device did not answer to a command") from err

                    logger.debug("exiting wait_for_message ...")
                    return None

    def send_message(self, message: HubMessage, keep: Callable[..., bool] = None):
        """
        Serializes a message and sends it to the interface, without waiting
        for an answer. Optionally, you can update the message queue filter
        if you need to wait for specific messages after the message is sent.

        :param Message message: Message to send
        :param keep: Message queue filter function
        """
        if not self.__opened and self.__out_messages.empty():
            raise WhadDeviceDisconnected()

        # Set message queue filter
        if keep is not None:
            self.set_queue_filter(keep)

        # Enqueue message to transmit to the interface
        self.__in_messages.put(message)


    def send_command(self, command: HubMessage, keep: Callable[..., bool] = None):
        """
        Sends a command and awaits a specific response from the device.
        WHAD commands usualy expect a CmdResult message, if `keep` is not
        provided then this method will by default wait for a CmdResult.

        :param Message command: Command message to send to the device
        :param keep: Message queue filter function (optional)
        :returns: Response message from the device
        :rtype: Message
        """
        # If a queue filter is not provided, expect a default CmdResult
        try:
            if keep is None:
                self.send_message(command, message_filter(CommandResult))
            else:
                self.send_message(command, keep)
        except WhadDeviceError as error:
            # Device error has been triggered, it looks like our device is in
            # an unspecified state, notify user.
            logger.debug("WHAD device in error while sending message: %s", error)
            raise error

        try:
            result = self.wait_for_message(self.__timeout, command=True)
        except WhadDeviceTimeout as timedout:
            # Forward exception
            raise timedout

        # Log message
        logger.debug("Command result: %s", result)

        return result

    ##
    # Interface management
    ##

    ######################################
    # Generic discovery
    ######################################

    def on_discovery_msg(self, message):
        """
        Method called when a discovery message is received. If a connector has
        been associated with the device, forward this message to this connector.
        """

        # Forward everything to the connector, if any
        if self.__connector is not None:
            self.__connector.on_discovery_msg(message)

    def has_domain(self, domain) -> bool:
        """Checks if device supports a specific domain.

        :param Domain domain: Domain
        :returns: True if domain is supported, False otherwise.
        :rtype: bool
        """
        if self.__info is not None:
            return self.__info.has_domain(domain)

        # No info available on device, domain is not supported by default
        return False


    def get_domains(self) -> dict:
        """Get device' supported domains.

        :returns: list of supported domains
        :rtype: list
        """
        if self.__info is not None:
            return self.__info.domains

        # No domain discovered yet
        return {}


    def get_domain_capability(self, domain):
        """Get a device domain capabilities.

        :param Domain domain: Target domain
        :returns: Domain capabilities
        :rtype: DeviceDomainInfoResp
        """
        if self.__info is not None:
            return self.__info.get_domain_capabilities(domain)

        # No capability if not discovered
        return 0

    def get_domain_commands(self, domain):
        """Get a device supported domain commands.

        :param Domain domain: Target domain
        :returns: Bitmask of supported commands
        :rtype: int
        """
        if self.__info is not None:
            return self.__info.get_domain_commands(domain)

        # No supported commands by default
        return 0


    def send_discover_info_query(self, proto_version=0x0100):
        """
        Sends a DeviceInfoQuery message and awaits for a DeviceInfoResp
        answer.
        """
        def discover_info_query_sent(message, status):
            print("Discover info query sent !")

        logger.info("preparing a DeviceInfoQuery message")
        msg = self.__hub.discovery.create_info_query(proto_version)
        msg.callback(discover_info_query_sent)
        return self.send_command(
            msg,
            message_filter(InfoQueryResp)
        )


    def send_discover_domain_query(self, domain):
        """
        Sends a DeviceDomainQuery message and awaits for a DeviceDomainResp
        answer.
        """
        logger.info("preparing a DeviceDomainInfoQuery message")
        msg = self.__hub.discovery.create_domain_query(domain)
        return self.send_command(
            msg,
            message_filter(DomainInfoQueryResp)
        )

    def discover(self):
        """
        Performs device discovery (synchronously).

        Discovery process asks the device to provide its description, including
        its supported domains and associated capabilities. For each domain we
        then query the device and get the list of supported commands.
        """
        if not self.__discovered:
            # We send a DeviceInfoQuery message to the device and expect a
            # DeviceInfoResponse in return.
            resp = self.send_discover_info_query()

            # If we have an answer, process it.
            if resp is not None:

                # Ensure response is the one we expect
                assert isinstance(resp, InfoQueryResp)

                # Save device information
                self.__info = WhadDeviceInfo(
                    resp
                )

                # Parse DeviceInfoResponse
                #device_info = self.hub.parse(resp)

                # Update our ProtocolHub version to the device version
                self.__hub = ProtocolHub(resp.proto_min_ver)

                # Query device domains
                logger.info("query supported commands per domain")
                for domain in self.__info.domains:
                    resp = self.send_discover_domain_query(domain)
                    self.__info.add_supported_commands(
                        resp.domain,
                        resp.supported_commands
                    )

                # Mark device as discovered
                logger.info("device discovery done")
                self.__discovered = True

                # Switch to max transport speed
                logger.info("set transport speed to %d", self.info.max_speed)
                self.change_transport_speed(
                    self.info.max_speed
                )
            else:
                logger.error("device is not ready !")
                raise WhadDeviceNotReady()

    def reset(self):
        """Reset device
        """
        logger.info("preparing a DeviceResetQuery message")
        msg = self.__hub.discovery.create_reset_query()
        return self.send_command(
            msg,
            message_filter(DeviceReady)
        )

class VirtualInterface(Interface):
    """
    Virtual interface implementation.
    
    This variant of the base Interface class provides a way to emulate an interface
    compatible with WHAD. This emulated compatible interface is used as an adaptation
    layer between WHAD's core and third-party hardware that does not run a WHAD-enabled
    firmware.
    """
    def __init__(self, index: int = None):
        self._dev_type = None
        self._dev_id = None
        self._fw_author = None
        self._fw_url = None
        self._fw_version = (0, 0, 0)
        self._dev_capabilities = {}
        self.__lock = Lock()
        super().__init__(index)

    def send_message(self, message, keep=None):
        """Send message to host.
        """
        with self.__lock:
            super().set_queue_filter(keep)
            self._on_whad_message(message)

    def _on_whad_message(self, message):
        """TODO: associate callbacks with classes ?
        """
        category = message.message_type
        message_type = message.message_name

        callback_name = f"_on_whad_{category}_{message_type}"
        if hasattr(self, callback_name) and callable(getattr(self, callback_name)):
            getattr(self, callback_name)(message)
        else:
            logger.info("unhandled message: %s", message)
            self._send_whad_command_result(ResultCode.ERROR)

    def _on_whad_discovery_info_query(self, message):
        major, minor, revision = self._fw_version
        msg = self.hub.discovery.create_info_resp(
            DeviceType.VirtualDevice,
            self._dev_id,
            0x0100,
            0,
            self._fw_author,
            self._fw_url,
            major, minor, revision,
            [domain | (capabilities[0] & 0xFFFFFF) for domain, capabilities in self._dev_capabilities.items()]
        )
        self._send_whad_message(msg)

    def _on_whad_discovery_domain_query(self, message):
        # Compute supported commands for domain
        commands = 0
        supported_commands = self._dev_capabilities[message.domain][1]
        for command in supported_commands:
            commands |= (1 << command)

        # Create a DomainResp message and send it
        msg = self.hub.discovery.create_domain_resp(
            message.domain,
            commands
        )
        self._send_whad_message(msg)


    def _send_whad_message(self, message):
        self.put_message(message)

    def _send_whad_command_result(self, code):
        msg = self.hub.generic.create_command_result(code)
        self._send_whad_message(msg)
